<?php
/**
 * @package     Module VEL Checker for Joomla
 * @version     1.1.6
 * @author      Stergios Zgouletas <info@web-expert.gr> / WEB EXPERT SERVICES LTD
 * @link        http://www.web-expert.gr
 * @copyright   Copyright (C) 2019-2020 WEB EXPERT SERVICES LTD. All Rights Reserved
 * @license     http://www.gnu.org/licenses/gpl-2.0.html GNU/GPL
 */
defined( '_JEXEC' ) or die( 'Restricted access' );

class modVelcheckerHelper
{
	//Get/Write/Reset the Ignore List with Reported IDs
	public static function ignoreList($uid=0)
	{	
		$list=dirname(__FILE__).'/.ignorelist';
		$listIDs=array();
		if(JFile::exists($list))
		{
			if($uid==-1)
			{
				//reset
				JFile::write($list,'');
			}
			else
			{
				//get
				$data=@file_get_contents($list);
				if($data!='') $listIDs=array_unique(array_filter(explode("\n",$data),'is_numeric'));
			}
		}
		
		//write
		if($uid>0 && !in_array($uid,$listIDs))
		{
			$listIDs[]=$uid;
			JFile::write($list,implode("\n",$listIDs));
		}
		
		return $listIDs;
	}
	
	//Get report, this functions searches for Vulnerable extensions
	public static function getReport($params,$ignoreList=array())
	{		
		$app=JFactory::getApplication();
		$VELdata=self::getList();
		
		if(isset($VELdata['error'])) //ability to call the helper from CLI
		{
			if(php_sapi_name() !== 'cli') $app->enqueueMessage($VELdata['error'], 'error');
			return array('error'=>$VELdata['error']);
		}
		
		#Debug
		if($params->get('debug',0))
		{
			echo 'Vel Data Rows:<pre><code>'.count($VELdata).'</code></pre><br/>';
		}
		
		$extraRows=array();
		foreach($VELdata as $ext)
		{					
			//find in the description LIST of extensions: ExtName: x.x.x
			if(preg_match_all("/(.*):\s((?:[0-9]+\.?)+)/i", strip_tags(html_entity_decode($ext['description'])), $matches))
			{
				foreach($matches[1] as $idx=>$extname)
				{				
					if(JString::strlen($extname)<3 || JString::strpos($matches[2][$idx],'.')===false) continue;
					$newRow=$ext;
					$newRow['title']=$extname.','.$matches[2][$idx].','.$ext['title'];
					$extraRows[]=$newRow;
				}
			}
		}
		
		#Debug
		if($params->get('debug',0))
		{
			echo 'Extra Rows Detected:<pre><code>'.count($extraRows).'</code></pre><br/>';
		}
		
		JPluginHelper::importPlugin('system');
		$dispatcher = class_exists('JEventDispatcher')? JEventDispatcher::getInstance() : JDispatcher::getInstance();
		
		$overall=array('indager'=>array(),'installed'=>array(),'stats'=>array());
		//stats
		$overall['stats']['indanger']['items']=0;
		$overall['stats']['indanger']['ignored']=0;	
		$overall['stats']['installed']['items']=0;
		$overall['stats']['installed']['ignored']=0;	
		
		$listVEL=$VELdata+$extraRows;
		$debugID=$app->input->get('debugid',0);
		
		foreach($listVEL as $ext)
		{		
			if($debugID)
			{
				$params->set('debug',$ext['id']==$debugID?1:0);
			}
			
			//Ignore very old reports
			$hideTime=strtotime($ext['modified'].' +'.$params->get('oldermonths',36).' month');
			if(!empty($ext['modified']) && $hideTime!==false && strtotime('now')>$hideTime)
			{
				if($params->get('debug',0))  echo 'HIDE: '.$ext['title'].' '.$ext['modified'].'<br/>';
				continue; //ingore to old
			}
			
			//Format: name,versions,reason
			$vars=explode(',',$ext['title']);
			$ext['reason']=end($vars);
			$accurancy=1;
			$versions=array();
			
			if(preg_match_all("/((?:[0-9]+\.?)+)/i", $ext['title'], $matches))
			{
				$matches=array_unique($matches[1]);
				$versions=array_filter($matches,function($v){return strpos($v,'.')!==false;});
				if(count($versions)) $accurancy+=2;
			}
			
			if(!count($versions) && preg_match_all("/((?:[0-9]+\.?)+)/i", $ext['description'], $matches))
			{
				$matches=array_unique($matches[1]);
				$versions=array_filter($matches,function($v){return strpos($v,'.')!==false;});
				
				if(count($versions)) $accurancy+=1;
			}
			
			if(!count($versions)) 
			{
				$versions=array('0.0.0'); //any
				$accurancy=0;
			}
			
			if(isset($ext['vulnerable_version']) && !in_array($ext['vulnerable_version'],$versions))
			{
				$versions[]=$ext['vulnerable_version'];
				$accurancy+=1;	
			}
			
			if(isset($ext['start_version']) && !in_array($ext['start_version'],$versions))
			{
				$versions[]=$ext['start_version'];
				$accurancy+=1;	
			}
			
			//remove the 0.0.0 if at least a version has been found
			if(count($versions)>1 && ($key = array_search('0.0.0', $versions)) !== false)
			{
				unset($versions[$key]);
			}
			
			//Detect Extension Name
			//Format EXTENSION NAME VERSOION_NUMBER. Extension name ends when the version number starts
			$versionPos=JString::strpos($vars[0],$versions[0]);
			if($versionPos!==false)
			{
				$vars[0]=JString::substr($vars[0],0,$versionPos);
			}			
			$potentialExtName=JString::trim($vars[0]);

			$query=array();
			$query['versions']=$versions;
			
			//If install_data exist, then use this data
			if(isset($ext['install_data']))
			{
				if(isset($ext['install_data']['name']))
				{
					$potentialExtName=JString::trim($ext['install_data']['name']);
					$accurancy++;
				}
				if(isset($ext['install_data']['type']))
					$query['type']=JString::strtolower($ext['install_data']['type']);
				if(isset($ext['install_data']['group']) && !empty($ext['install_data']['group']))
					$query['group']=JString::strtolower($ext['install_data']['group']);
			}
			
			# Start Extension Name FIX
			#========================================================#
			$potentialExtName=JString::strtolower($potentialExtName);

		   //Format: Ext_name {KEYWORD} vendor/company Example: Magiczoomplus for Virtuemart
			$breakKeywords=array(' for ',' by ');
			foreach($breakKeywords as $keyword)
			{
				$versionPos=JString::strpos($potentialExtName,$keyword);
				if($versionPos!==false)
				{
					$potentialExtName=JString::substr($potentialExtName,0,$versionPos);
				}
			}
			
			//fix extension name
			$remove=array(' extension','codels ','joomlaworks ',' previous',' versions',' pre',' version',' v',' for',' by','  cart',' pro',' free','joomla ',' joomla',' plugin',' older',' than','!','-',' library',' module',' template',' component',' package',' forum',' installer');
			$potentialExtName=str_replace($remove,'',$potentialExtName); //name
			$potentialExtName = preg_replace('!\s+!', ' ', $potentialExtName); //double+ space with signle
			
			//This bad solution, but the incoming data SUCKS.
			$potentialExtName=str_replace('all regular labs extensions with editor buttons','regular labs',$potentialExtName); //Regular labs
			$potentialExtName=str_replace(array('akeeba backup','akeeba cms update'),'com_akeeba',$potentialExtName); //Akeeba
			$potentialExtName=str_replace('akeeba loginguard','loginguard',$potentialExtName); //Akeeba LoginGuard
			$potentialExtName=str_replace('booking calendar','bookingcalendar',$potentialExtName); //Causes false positive
						
			$potentialExtName=preg_replace("/\(([^()]*+|(?R))*\)/","", $potentialExtName); //remove parenthis text like (pro version)
			$query['search']=$potentialExtName;
			# END Extension Name FIX
			#========================================================#
					
			//For developers: System Plugin Events
			$dispatcher->trigger('onBeforeVelExtSearch',array(&$query,&$ext));
			$results=self::searchExtension($query);
			$dispatcher->trigger('onAfterVelExtSearch',array(&$results,&$ext));

			#Debug
			if($params->get('debug',0))
			{
				echo '<pre>ID:'.$ext['id'].' Title:<code>'.$ext['title'].' | Ext. Name:'.$potentialExtName.'</code></pre><br/>';
				echo 'Vel Data:<pre><code>'.print_r($ext,true).'</code></pre><br/>';
				echo 'Debug:<pre><code>'.print_r($results['debug'],true).'</code></pre><br/>';
				echo 'Result:<pre><code>'.print_r($results['data'],true).'</code></pre><br/><hr/>';
			}
			
			if(count($results['data']))
			{
				if(isset($results['debug']['input']['type']) || isset($results['debug']['input']['folder'])) $accurancy++;
				if($results['total']<3) $accurancy++;
				
				foreach($results['data'] as $extID=>$extData)
				{				
					$extData['uid']=$extID+$ext['id'];
					if(in_array($extData['uid'],$ignoreList))
					{
						$extData['ignored']=true;
						$extData['indanger']? $overall['stats']['indanger']['ignored']++ : $overall['stats']['installed']['ignored']++;
					}
					
					$ext['vulnerable_versions']=$results['debug']['input']['versions'];
					$extData['velinfo']=$ext;
					$extData['accurancy']=$accurancy;
					//$extData['debug']=$results['debug'];
					$extData['indanger']? $overall['indanger'][]=$extData : $overall['installed'][]=$extData;
				}
			}
		}
		
		$overall['stats']['indanger']['items']=count($overall['indanger'])-$overall['stats']['indanger']['ignored'];
		$overall['stats']['installed']['items']=count($overall['installed'])-$overall['stats']['installed']['ignored'];
		
		return $overall;
	}
	
	//Search Extension based on query values.
	protected static function searchExtension($query,$inDangerOnly=false)
	{
		$noResult=array('data'=>array(),'total'=>0,'debug'=>array('input'=>$query));
		if(empty($query['versions']) || empty($query['search'])) return $noResult;
	
		$db = JFactory::getDBO();
		$search=array();

		$query['search']=JString::trim($query['search']);
		
		//Build extension name with the prefixes
		$prefixes=array('component'=>'com_','module'=>'mod_','plugin'=>'plg_','package'=>'pkg_');
		if(!isset($query['type']) && JString::strpos($query['search'],'_')!==false)
		{
			foreach($prefixes as $type=>$prefix)
			{
				if(JString::substr($query['search'],0,4)==$prefix)
				{
					$query['type']=$type;
					break;
				}
			}
		}
		
		if(isset($query['type']))
		{
			$prefix='';
			if(array_key_exists($query['type'],$prefixes))
			{
				$prefix=$prefixes[$query['type']];
				$search['type']='`type`='.$db->quote($query['type']);
			}
				
			if($query['type']=='plugin')
			{
				$prefix=''; //plugins do not use prefix
				if(JString::strpos($query['search'],'plg_')!==false)
				{
					#example: plg_type_extename
					preg_match('/plg_(.*)_(.*)$/',$query['search'],$matches);
					if(isset($matches[1]) && !empty($matches[1]))
					{
						$query['search']=str_replace($matches[1].'_','',$query['search']);
						if(!isset($search['folder'])) $search['folder']='`folder`='.$db->quote($matches[1]);
					}
				}
			}
						
			if($prefix!='' && JString::substr($query['search'],0,4)!=$prefix)
			{
				$query['search']=$prefix.$query['search'];
			}
			
			#example: plg_extename
			$query['search']=str_replace('plg_','',$query['search']);
		}
		
		if(isset($query['group']))
		{
			$search['type']='`type`='.$db->quote('plugin');
			$search['folder']='`folder`='.$db->quote($query['group']);
		}
				
		if(JString::strlen($query['search'])<3 || JString::strlen($query['search'])>30)
		{
			$noResult['debug']['input']=$query;
			return $noResult; //too short or too big does not seems to be extension name		
		}
		
		//Check if there is space
		if(JString::strpos($query['search'],' ')!==false)
		{
			//If has more than 3 spaces probably it's phraze and not extension name
			$parts=explode(' ',$query['search']);
			if(count($parts)>3)
			{
				$noResult['debug']['input']=$query;
				return $noResult;
			}
			
			$searchNames=array();
			$searchNames[]='`element`='.$db->quote(str_replace(' ','',$query['search']));
			$searchNames[]='`element`='.$db->quote(str_replace(' ','_',$query['search']));
			$last=end($parts);
			$medi=JString::trim(str_replace(array($parts[0],$last),'',$query['search']));
			
			#Example: community builder
			#match 1st medium last word
			$searchNames[]='(`element` LIKE '.$db->quote($db->escape($parts[0]).'%').' AND `element` LIKE '.$db->quote('%'.$db->escape($medi).'%').' AND `element` LIKE '.$db->quote('%'.$db->escape($last)).')';
			#match 1st and last word
			$searchNames[]='(`element` LIKE '.$db->quote($db->escape($parts[0]).'%').' AND `element` LIKE '.$db->quote('%'.$db->escape($last)).')';
			#match medium last word
			//$searchNames[]='(`element` LIKE '.$db->quote('%'.$db->escape($medi).'%').' AND `element` LIKE '.$db->quote('%'.$db->escape($last)).')';
			#match 1st  medium word
			//$searchNames[]='(`element` LIKE '.$db->quote($db->escape($parts[0]).'%').' AND `element` LIKE '.$db->quote('%'.$db->escape($medi).'%').')';
			
			#This causes many false positives
			//if(JString::strlen($last)>5) $searchNames[]='(`element` LIKE '.$db->quote('%'.$db->escape($last)).')';
			
			$search[]='('.implode(' OR ',$searchNames).')';
		}
		else
		{
			$search[]='(`element`='.$db->quote($query['search']).' OR `element` LIKE '.$db->quote('%'.$db->escape($query['search'])).')';	
		}
		
		
		static $cache=array();
		//skip query if we already searched before with same terms
		$cacheKey=md5(implode('',$search));
		if(isset($cache[$cacheKey]) && is_array($cache[$cacheKey])) 
		{
			$extensions=$cache[$cacheKey];
		}
		else
		{
			$db->setQuery('SELECT * FROM `#__extensions` WHERE '.implode(' AND ',$search).' ORDER BY `type` ASC LIMIT 0,4;');
			$extensions=$db->loadObjectList();
			$cache[$cacheKey]=$extensions;
		}
				
		$results=array('data'=>array(),'total'=>count($extensions),'debug'=>array());
		$debugVersions=array();
		$compoIgnorePlugins=true; //if component found ignore plugins
		$ignorePlg=false;
		foreach($extensions as $ext)
		{
			if(!$ext->extension_id) continue;
			
			if($compoIgnorePlugins && $ext->type=='component') $ignorePlg=true;
			if($ignorePlg && $ext->type=='plugin') continue;
			
			$xml = json_decode($ext->manifest_cache);
			$inDanger=!count($query['versions']);
			foreach($query['versions'] as $idx=>&$v)
			{
				if(JString::strpos($v,'.')===false)
				{
					unset($query['versions'][$idx]);
					continue;
				}
				
				$v=str_replace('..','.',$v); //fix version 3..3.9
				$v=str_replace('x','0',$v); //fix version 2.2.x
				$v=trim($v);
				
				if(JString::substr($v,-1)=='.')
				{
					$vp=explode('.',$v);				
					if(count($vp)<3)
						$v.='0'; //fix version format "d.d."
					else
						$v=trim($v,'.'); //fix version format "d.d.d."
				}
				
				//Exlcude dates like 2014.0
				$vp=explode('.',$v,2);
				if((int)$vp[0]>=1970)
				{
					unset($query['versions'][$idx]);
					continue;
				}
						
				#Debug
				$debugVersions[]=$ext->name.' ['.$query['search'].'] '.$v.' Ext:'.$xml->version.' '.(int)version_compare($xml->version,$v,'le').'<br/>';
				if($v=='0.0.0' || version_compare($xml->version,$v,'le'))
				{
					$inDanger=true;
					break;
				}
			}
			if($inDangerOnly && !$inDanger) continue;
			$results['data'][$ext->extension_id]=array('id'=>$ext->extension_id,'name'=>$ext->name,'element'=>$ext->element,'type'=>$ext->type,'folder'=>$ext->folder,'cversion'=>$xml->version,'indanger'=>$inDanger);
		}
				
		$results['debug']=array(
			'input'=>$query,
			'sql_filters'=>implode(' AND ',$search),
			'total_found_db'=>count($extensions),
			'version_matched'=>$inDanger,
			'versions_check'=>$debugVersions,			
		);
				
		$cache[$cacheKey]=$results;
		return $results;
	}

	//It's appropriate to cache the file to avoid multiple requests that may cause sever blocks
	public static function getList()
	{
		//use File Cache Extension
		$cache = JFactory::getCache('mod_velchecker_list','callback','file',array('storage'=>'file'));
		$cache->setCaching(1);
		$cache->setLifeTime(3*60); //3 hours
		
		$url='https://extensions.joomla.org/index.php?option=com_vel&format=json&_c='.date('Y-m-d'); //use date to clear cache every day
		$velData= $cache->call(array(__CLASS__,'getRemote'),$url);

		$velData=@json_decode($velData,true);
		if($velData===NULL || !$velData['success'])
		{
			$cache->clean();
			return array('error'=>'Unable to retrieve VEL data');
		}
		
		$velDataHash = str_replace(array(" ","\t","\r","\n"),"",json_encode($velData['data']));
		$velDataHash = hash("sha256",$velDataHash);
		
		$url='https://extensions.joomla.org/index.php?option=com_vel&format=json&task=verify&_c='.date('Y-m-d');
		$verifyData= $cache->call(array(__CLASS__,'getRemote'),$url);
		$verifyData=@json_decode($verifyData,true);

		if($verifyData===NULL || !$verifyData['success'])
		{
			$cache->clean();
			return array('error'=>'Unable to retrieve VEL Verify data');
		}
		
		return $verifyData['data']==$velDataHash?$velData['data']['items']:array('error'=>'Verification failed, invalid VEL data');
	}
	

	public static function getRemote($url)
	{
		$user_agent = isset($_SERVER['HTTP_USER_AGENT']) ? $_SERVER['HTTP_USER_AGENT'] : 'Mozilla/5.0 (JMOD_VELCHECKER 1.1.4)';
		// Create a stream
		$opts = [
			"http" => [
				"method" => "GET",
				"header" => "Accept-language: en\r\n" .
				              "User-Agent: ".$user_agent."\r\n",
			]
		];

		$context = stream_context_create($opts);
		return @file_get_contents($url, false, $context);		
	}	
}